// -*- Mode:C++ -*-
//Header:
//File: WorldModel.h
//Title: WorldModel.h
//Author: NODA, Itsuki
//Date: 2001/07/27
//

//ModifyHistory:
// 2001/07/27: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_WorldModel_h_
#define _SPlayer_WorldModel_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"
#include "SensorInfo.h"
#include "BasePlayer.h"
#include "estimatePos.h"
#include "FieldState.h"

/* <<< [Bun-0.2] <<< */
#include "Newton2dKFilter.h"
/* >>> [Bun-0.2] >>> */

namespace Soccer {
   
   using namespace Itk ;

   //======================================================================
   // class WorldModel

   /*--------------------*/
   /**
    *
    */

   class WorldModel : public WithDescriber {

	 //------------------------------------------------------------
	 // WorldModel:: local classes
      public:
	 class Action ;

      public:
	 class State ;

	 /*--------------------*/
	 /**
	  *
	  */

	 typedef FieldState::Object Object ;
      
	 //------------------------------------------------------------
	 // WorldModel:: member

	 class DummyMutex {
	    public: void lock() {} ;
	    public: void unlock() {} ;
	 } ;

      public:
	 Int worktime ;
	 DummyMutex mutex ;

	 Flt decayConf ;
	 Flt decayPlayerVel ;
	 Flt decayBallVel ;

	 //----------------------------------------
	 // WorldModel:: state, ball, self
      public:
	 static const Int stateN = 64 ;
	 State * state_ ;

	 ITK_DEF_ACCESS(State & state(Int t), { 
	    if(t < 0) 
	       return state_[0] ;
	    else 
	       return state_[t % stateN]; 
	 }) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Object & ball(Int t), { return state(t).ball() ; });

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Object & self(Int t), { return state(t).self() ; });

	 /*--------------------*/
	 /**
	  *
	  */
	 /* <<< [Bun-0.2] <<< */
      public:
	 ITK_DEF_ACCESS3(BasePlayer *, & me(), { return me_ ; }) ;
	 BasePlayer * me_ ;
	 /* >>> [Bun-0.2] >>> */

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Action & action(Int t), {
	    if(t < 0)
	       return action_[0] ;
	    else
	       return action_[t % stateN]; 
	 }) ;
	 Action * action_ ;

	 //----------------------------------------
	 // WorldModel:: server and misc parameters
      public:
	 Flt useSensorCutoff ;
	 Flt mergeSensorRate ;
	 Flt playerSpeedMax ;
	 Flt shortRangeThreshold ;
	 Flt ballSpeedMax ;
	 Flt dashPowerRate ;
	 Flt inertia ;
	 Flt neckMax ;
	 Flt kickableMargin ;
	 Flt kickPowerRate ;
	 Flt playerSize ;
	 Flt ballSize ;
	 Flt distQuantizeStep ;

	 /* <<< [Bun-0.3] <<< */
	 Flt playerRand ;	// player_rand param in server.conf
	 Flt ballRand ;		// ball_rand param in server.conf
	 Flt playerSigmaAcc ;	// variance of accel. of player
	 Flt ballSigmaAcc ;	// variance of accel. of ball
	 /* >>> [Bun-0.3] >>> */

	 /* <<< [Bun-0.2] <<< */

	 /*--------------------*/
	 /**
	  * Kalman Filter for Ball
	  */
      public:
	 ITK_DEF_ACCESS(Newton2dKFilter & 
			ballKFilter(), { return ballKFilter_ ; }) ;
	 Newton2dKFilter ballKFilter_ ;

	 /*--------------------*/
	 /**
	  * Kalman Filter for Self
	  */
      public:
	 ITK_DEF_ACCESS(Newton2dKFilter & 
			selfKFilter(), { return selfKFilter_ ; }) ;
	 Newton2dKFilter selfKFilter_ ;

	 /* >>> [Bun-0.2] >>> */

	 //------------------------------------------------------------
	 // WorldModel:: constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 /* <<< [Bun-0.2] <<< */
	 // WorldModel() {
	 WorldModel(BasePlayer * myplayer = ITK_NULLPTR) {
	    init(myplayer) ;
	 } 
	 /* >>> [Bun-0.2] >>> */

	 //------------------------------------------------------------
	 // WorldModel:: init

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 /* <<< [Bun-0.2] <<< */
	 void init(BasePlayer * myplayer = ITK_NULLPTR) { // !!! [Bun-0.2] !!!
	    state_ = new State[stateN] ;
	    action_ = new Action[stateN] ;

	    worktime = 0 ;

	    me() = myplayer ; // !!! [Bun-0.2] !!!

	    initServerParam() ;

	    /* <<< [Bun-0.2] <<< */

	    ballKFilter().init() ;
	    ballKFilter().setParam(decayBallVel, // alpha 
				   0.01, 	// sigma pos 
				   0.01, 	// sigma vel 
				   0.1,  	// sigma acc 
				   0.3);  	// sigma observ
	    ballKFilter().initMatrix() ;

	    selfKFilter().init() ;
	    selfKFilter().setParam(decayPlayerVel, // alpha 
				   0.01, 	// sigma pos 
				   0.01, 	// sigma vel 
				   0.1,  	// sigma acc 
				   0.3);  	// sigma observ
	    selfKFilter().initMatrix() ;

	    /* >>> [Bun-0.2] >>> */
				   
	 } ;
	 /* >>> [Bun-0.2] >>> */

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void initServerParam() {
	    decayConf = 0.5 ;

	    decayPlayerVel = 0.4 ;	// from server.conf
	    decayBallVel = 0.94 ;       // from server.conf

	    playerSpeedMax = 1.0 ; // from server.conf
	    ballSpeedMax = 2.7 ; // from server.conf
	    dashPowerRate = 0.006 ; // from server.conf
	    inertia = 5.0 ; // from server.conf
	    neckMax = 90.0 ; // from server.conf

	    kickableMargin = 0.7 ; // from server.conf
	    kickPowerRate = 0.027 ;  // from server.conf
	    playerSize = 0.3 ;  // from server.conf
	    ballSize = 0.085 ;  // from server.conf

	    useSensorCutoff = 1.0 ;  // from estimated value
	    mergeSensorRate = 0.5 ;  // ???
	    shortRangeThreshold = 3.0 ; // from server.conf ?

	    distQuantizeStep = 0.1 ; // from server.conf

	    /* <<< [Bun-0.3] <<< */
	    playerRand = 0.1 ;
	    ballRand = 0.05 ;
	    playerSigmaAcc = 0.1 ;
	    ballSigmaAcc = 0.1 ;
	    /* >>> [Bun-0.3] >>> */
	 } ;


	 //------------------------------------------------------------
	 // WorldModel:: estState, estBall, estSelf
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual State & estState(Int t, Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Object & estBall(Int t, Bool lockp = True) { 
	    return estState(t,lockp).ball() ; 
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Object & estSelf(Int t, Bool lockp = True) { 
	    return estState(t,lockp).self() ; 
	 } ;

	 //------------------------------------------------------------
	 // WorldModel:: estimate1
	 //   estimate state(t) from state(t-1)
	 //   suppose state(t-1) is calculated
	 //   return state(t)
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual State & estimate1(Int t, Bool lockp = True) ;
	    
	 //------------------------------------------------------------
	 // WorldModel:: put sensored data
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedData(Int t, const FieldState & fs, 
			      Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataSelf(Int t, const FieldState & fs, 
				  Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataSelfPos(Int t, const FieldState & fs, 
				     State & cstate,
				     Object & self,
				     Bool & useSensor) ; 
				     
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataSelfVel(Int t, const FieldState & fs, 
				     State & cstate,
				     Object & self,
				     Bool & useSensor) ; 

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataBall(Int t, const FieldState & fs, 
				  Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataBallPos(Int t, const FieldState & fs, 
				     State & cstate,
				     Object & ball,
				     Bool & useSensor, 
				     Bool & shortRange) ;
				     
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mergeSensedDataBallVel(Int t, const FieldState & fs, 
				     State & cstate,
				     Object & ball,
				     Bool & useSensor, 
				     Bool & shortRange) ;

	 /* <<< [Bun-0.2] <<< */

	 /*--------------------*/
	 /**
	  * 
	  * Kalman Filter Version
	  */
      public:
	 void mergeSensedDataSelf2(Int t, const FieldState & fs, 
				   Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  * Kalman Filter Version
	  */
      public:
	 void mergeSensedDataBall2(Int t, const FieldState & fs, 
				   Bool lockp = True) ;
	 /* >>> [Bun-0.2] >>> */

	 //------------------------------------------------------------
	 // WorldModel:: estimateFore
	 //   estimate state(t) from state(t-1)
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void estimateFore(State & cstate, 
				   const State & pstate, 
				   const Action & paction) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void estimateSelfFore(State & cstate, 
				       const State & pstate, 
				       const Action & paction) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void estimateBallFore(State & cstate, 
				       const State & pstate, 
				       const Action & paction) ;

	 /* <<< [Bun-0.2] <<< */

	 /*--------------------*/
	 /**
	  *
	  * Kalman Filter version
	  */
      public:
	 virtual void estimateSelfFore2(State & cstate, 
					const State & pstate, 
					const Action & paction) ;

	 /*--------------------*/
	 /**
	  *
	  * Kalman Filter version
	  */
      public:
	 virtual void estimateBallFore2(State & cstate, 
					const State & pstate, 
					const Action & paction) ;
	 /* >>> [Bun-0.2] >>> */

	 //--------------------
	 // WorldModel:: util. for kick ball
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Flt kickPowerFactor(Flt rdist, Flt rdir) {
	    return kickPowerRate * (1.0 
				    - (0.25 * deg2rad(rdir) / M_PI)
				    - (0.25 * rdist / kickableMargin)) ;
	 } ;

	 //------------------------------------------------------------
	 // WorldModel:: estimateBacks
	 //   estimate state(t) from state(t+1)

	 //------------------------------
	 // WorldModel:: estimateBack action by ball vel
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateBallBackActionByBallVel(Action &caction,
						      State & nstate,
						      const State & cstate) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateBallBackActionByBallVel(Action &caction,
						      const Vec & nbvel,
						      const State & cstate) ;

	 //------------------------------
	 // WorldModel:: estimateBack action by ball pos
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateBallBackActionByBallPos(Action &caction,
						      State & nstate,
						      const State & cstate) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateBallBackActionByBallPos(Action &caction,
						      const Vec & nbpos,
						      const State & cstate) ;

	 //============================================================
	 // WorldModel:: class Action
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class Action : public WithDescriber {

	       //--------------------------------------------------
	       // WorldModel::Action:: variables
	    public:
	       Bool varidp_ ;
	       ITK_DEF_ACCESS(Bool & varidp(), { return varidp_ ; }) ;

	    public:
	       Int time_ ;
	       ITK_DEF_ACCESS(Int & time(), { return time_ ; }) ;

	    public:
	       Bool dashp_ ;
	       ITK_DEF_ACCESS(Bool & dashp(), { return dashp_ ; }) ;
	    public:
	       Flt dashPower_ ;
	       ITK_DEF_ACCESS(Flt & dashPower(), { return dashPower_ ; }) ;

	    public:
	       Bool turnp_ ;
	       ITK_DEF_ACCESS(Bool & turnp(), { return turnp_ ; }) ;
	    public:
	       Flt turnPower_ ;
	       ITK_DEF_ACCESS(Flt & turnPower(), { return turnPower_ ; }) ;

	    public:
	       Bool kickp_ ;
	       ITK_DEF_ACCESS(Bool & kickp(), { return kickp_ ; }) ;
	    public:
	       Flt kickPower_ ;
	       ITK_DEF_ACCESS(Flt & kickPower(), { return kickPower_ ; }) ;
	    public:
	       Flt kickDir_ ;
	       ITK_DEF_ACCESS(Flt & kickDir(), { return kickDir_ ; }) ;
	       
	    public:
	       Bool headp_ ;
	       ITK_DEF_ACCESS(Bool & headp(), { return headp_ ; }) ;
	    public:
	       Flt headPower_ ;
	       ITK_DEF_ACCESS(Flt & headPower(), { return headPower_ ; }) ;

	       //--------------------------------------------------
	       // WorldModel::Action:: clear
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void clear() {
		  varidp() = False ;
		  time() = 0 ;
		  dashp() = False ;
		  turnp()= False ;
		  kickp() = False ;
		  headp() = False ;
	       } ;

	       //--------------------------------------------------
	       // WorldModel::Action:: setup
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setup(Int t = 0) {
		  clear() ;
		  varidp() = True ;
		  time() = t ;
	       } ;

	       //--------------------------------------------------
	       // WorldModel::Action:: set
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setDash(Flt power) {
		  dashp() = True ;
		  dashPower() = power ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setTurn(Flt power) {
		  turnp() = True ;
		  turnPower() = power ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setKick(Flt power, Flt dir) {
		  kickp() = True ;
		  kickPower() = power ;
		  kickDir() = dir ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setHead(Flt power) {
		  headp() = True ;
		  headPower() = power ;
	       } ;

	       //--------------------------------------------------
	       // WorldModel::Action:: describe
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       virtual void describe(ostream& ostr, 
				     const Bool detailp = True) const {
		  ostr << "#Action[time=" << time() << ":" ;

		  Bool bcomp = False ;

		  if(dashp()) {
		     if(bcomp) ostr << "," ;
		     ostr << "dash(" << dashPower() << ")" ;
		     bcomp = True ;
		  }

		  if(turnp()) {
		     if(bcomp) ostr << "," ;
		     ostr << "turn(" << turnPower() << ")" ;
		     bcomp = True ;
		  }

		  if(kickp()) {
		     if(bcomp) ostr << "," ;
		     ostr << "kick(" << kickPower() << "," 
			  << kickDir() << ")" ;
		     bcomp = True ;
		  }

		  if(!bcomp) ostr << "nop()" ;
		  
		  if(headp()) {
		     ostr << "," 
			  << "head(" << headPower() << ")" ;
		  }
		  
		  ostr << "]" ;
		  
		  if(detailp) ostr << endl ;
	       } ;
	 } ;

	 //============================================================
	 // WorldModel:: class State
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class State : public FieldState {

	       //----------------------------------------
	       // WorldModel::State:: enum Authority
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       enum Authority {
		  Auth_None,
		  Auth_Sensor,
		  Auth_Estimate,
		  Auth_Mixed,
		  Auth_Error,
		  Auth__Max,
	       } ;
	       
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       static const char * authorityName(Authority auth) {
		  switch(auth) {
		     case Auth_None : 	  return "none" ;
		     case Auth_Sensor :	  return "sensor" ;
		     case Auth_Estimate : return "estimate" ;
		     case Auth_Mixed :	  return "mixed" ;
		     case Auth_Error :    
		     default:
			                  return "error" ;
		  }
	       } ;
	       
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       static Authority authorityByName(const char * name) {
		  for(Int auth = (Int)Auth_None ; auth < (Int)Auth__Max ; 
		      ++auth){
		     if(!strcmp(name,authorityName((Authority)auth))) {
			return (Authority)auth ;
		     }
		  } 
		  return Auth_Error ;
	       } ;

	       //----------------------------------------
	       // WorldModel::State:: member
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(TimeVal & 
			      updateTime(), { 
		  return updateTime_ ; }) ;
	       TimeVal updateTime_ ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Int & 
			      time(), { return time_ ; }) ;
	       Int time_ ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Authority & 
			      selfAuthority(), { 
		  return selfAuthority_; });
	       Authority selfAuthority_ ;

	       const char * selfAuthorityName() {
		  return authorityName(selfAuthority()) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Int & 
			      selfNoSensor(), {
		  return selfNoSensor_ ; }) ;
	       Int selfNoSensor_ ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Flt & 
			      selfSensorMixRatio(), {
		  return selfSensorMixRatio_ ;}) ;
	       Flt selfSensorMixRatio_ ;


	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Authority & 
			      ballAuthority(), { 
		  return ballAuthority_;});
	       Authority ballAuthority_ ;

	       const char * ballAuthorityName() {
		  return authorityName(ballAuthority()) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Int & 
			      ballNoSensor(), {
		  return selfNoSensor_ ; }) ;
	       Int ballNoSensor_ ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Flt & 
			      ballSensorMixRatio(), {
		  return ballSensorMixRatio_ ;}) ;
	       Flt ballSensorMixRatio_ ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Flt & 
			      myspeed(), { return myspeed_ ;}) ;
	       Flt myspeed_ ;

	       /* <<< [Bun-0.2] <<< */
	       //----------------------------------------
	       // WorldModel::State:: for Kalman Filter
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(ValMatrix & 
			      ballKFilterPf(), { return ballKFilterPf_ ;}) ;
	       ValMatrix ballKFilterPf_ ;
	       /* >>> [Bun-0.2] >>> */

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(ValMatrix & 
			      selfKFilterPf(), { return selfKFilterPf_ ;}) ;
	       ValMatrix selfKFilterPf_ ;
	       /* >>> [Bun-0.2] >>> */
	       
	       //----------------------------------------
	       // WorldModel::State:: reset
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void reset() {
		  time() = -1 ;
		  updateTime().update() ;

		  selfAuthority() = Auth_None ;
		  ballAuthority() = Auth_None ;
		  selfNoSensor() = 0 ;
		  ballNoSensor() = 0 ;
		  selfSensorMixRatio() = 0.0 ;
		  ballSensorMixRatio() = 0.0 ;

		  self().time().init() ;
		  self().posConf = 0 ;
		  self().velConf = 0 ;

		  ball().time().init() ;
		  ball().posConf = 0 ;
		  ball().velConf = 0 ;

		  myspeed() = -1.0 ;  // if negative the value is not set

		  /* <<< [Bun-0.2] <<< */
		  if(ballKFilterPf().sizeN() == 0) {
		     ballKFilterPf().alloc(Newton2dKFilter::dfltSize,
					   Newton2dKFilter::dfltSize) ;
		     selfKFilterPf().alloc(Newton2dKFilter::dfltSize,
					   Newton2dKFilter::dfltSize) ;
		  }
		  /* >>> [Bun-0.2] >>> */
	       } ;

	       //----------------------------------------
	       // WorldModel::State:: merge speed info
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void mergeMyspeed(Flt speed) ;

	 } ;
	 
   } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif
